import jwt

from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, BaseUser, AuthCredentials,
    UnauthenticatedUser)


class JWTUser(BaseUser):
    def __init__(self, username: str, token: str, payload: dict) -> None:
        self.username = username
        self.token = token
        self.payload = payload

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.username


class JWTAuthenticationBackend(AuthenticationBackend):

    def __init__(self, secret_key: str, algorithm: str = 'HS256', prefix: str = 'JWT', username_field: str = 'username'):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.prefix = prefix
        self.username_field = username_field

    @classmethod
    def get_token_from_header(cls, authorization: str, prefix: str):
        """
        Parses the Authorization header and returns only the token
        :param authorization:
        :return:
        """
        try:
            scheme, token = authorization.split()
        except ValueError:
            raise AuthenticationError('Could not separate Authorization scheme and token')
        if scheme.lower() != prefix.lower():
            raise AuthenticationError(f'Authorization scheme {scheme} is not supported')
        return token

    async def authenticate(self, request):
        if "Authorization" not in request.headers:
            return None
        auth = request.headers["Authorization"]
        token = self.get_token_from_header(authorization=auth, prefix=self.prefix)
        try:
            payload = jwt.decode(token, key=self.secret_key, algorithms=self.algorithm)
        except jwt.InvalidTokenError as e:
            raise AuthenticationError(str(e))

        return AuthCredentials(["authenticated"]), JWTUser(username=payload[self.username_field],
                                                           token=token,
                                                           payload=payload)


class JWTWebSocketAuthenticationBackend(AuthenticationBackend):

    def __init__(self, secret_key: str, algorithm: str = 'HS256', query_param_name: str = 'jwt',
                 username_field: str = 'username'):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.query_param_name = query_param_name
        self.username_field = username_field

    async def authenticate(self, request):
        if self.query_param_name not in request.query_params:
            return AuthCredentials(), UnauthenticatedUser()
        token = request.query_params[self.query_param_name]

        try:
            payload = jwt.decode(token, key=self.secret_key, algorithms=self.algorithm)
            print(payload)
        except jwt.InvalidTokenError as e:
            raise AuthenticationError(str(e))

        return AuthCredentials(["authenticated"]), JWTUser(username=payload[self.username_field],
                                                           token=token,
                                                           payload=payload)



"""High-level Action base class implementation."""

import uuid

from .utils import timestamp
from .model import Thing as ThingModel


class Action:
    """An Action represents an individual action on a thing."""

    name = ""

    def __init__(self, thing, input_, id_=uuid.uuid4().hex):
        """
        Initialize the object.
        id_ ID of this action, default uuid
        thing -- the Thing this action belongs to
        name -- name of the action
        input_ -- any action inputs
        """
        self.id = id_
        self.thing = thing
        self.input = input_
        self.href_prefix = ""
        self.href = f"/actions/{self.name}/{self.id}"
        self.status = "created"
        self.time_requested = timestamp()
        self.time_completed = None

    async def as_action_description(self):
        """
        Get the action description.
        Returns a dictionary describing the action.
        """
        description = {
            self.name: {
                "href": self.href_prefix + self.href,
                "timeRequested": self.time_requested,
                "status": self.status,
            },
        }

        if self.input is not None:
            description[self.name]["input"] = self.input

        if self.time_completed is not None:
            description[self.name]["timeCompleted"] = self.time_completed

        return description

    async def set_href_prefix(self, prefix):
        """
        Set the prefix of any hrefs associated with this action.
        prefix -- the prefix
        """
        self.href_prefix = prefix

    async def get_id(self):
        """Get this action's ID."""
        return self.id

    async def get_name(self):
        """Get this action's name."""
        return self.name

    async def get_href(self):
        """Get this action's href."""
        return self.href_prefix + self.href

    async def get_status(self):
        """Get this action's status."""
        return self.status

    async def get_thing(self):
        """Get the thing associated with this action."""
        return self.thing

    async def get_time_requested(self):
        """Get the time the action was requested."""
        return self.time_requested

    async def get_time_completed(self):
        """Get the time the action was completed."""
        return self.time_completed

    async def get_input(self):
        """Get the inputs for this action."""
        return self.input

    async def start(self):
        """Start performing the action."""
        self.status = "pending"
        await self.thing.action_notify(self)
        await self.perform_action()
        await self.finish()

    async def perform_action(self):
        """Override this with the code necessary to perform the action."""
        pass

    async def cancel(self):
        """Override this with the code necessary to cancel the action."""
        pass

    async def finish(self):
        """Finish performing the action."""
        self.status = "completed"
        self.time_completed = timestamp()
        await self.thing.action_notify(self)


class Rename(Action):
    name = "rename"

    async def perform_action(self):
        tuple = await ThingModel.get_or_create(uid=self.thing.id, title=self.thing.title)
        thing = tuple[0]
        await ThingModel.filter(uid=thing.uid).update(title=self.input["title"])



"""High-level Property base class implementation."""

from copy import deepcopy
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from .errors import PropertyError


class Property:
    """A Property represents an individual state value of a thing."""

    __slots__ = ['thing', 'name', 'value', 'metadata', 'href_prefix', 'href', 'media_type']

    def __init__(self, name, value, thing=None, metadata=None):
        """
        Initialize the object.
        thing -- the Thing this property belongs to
        name -- name of the property
        value -- Value object to hold the property value
        metadata -- property metadata, i.e. type, description, unit, etc.,
                    as a dict
        """
        self.thing = thing
        self.name = name
        self.value = value
        self.metadata = metadata if metadata is not None else {}
        self.href_prefix = ""
        self.href = f"/properties/{self.name}"
        self.media_type = "application/json"

        # Add the property change observer to notify the Thing about a property
        # change.
        self.value.on("update", lambda _: self.thing.property_notify(self))
        self.value.on("sync", lambda _: self.thing.property_notify(self))
        self.value.on("update", lambda _: self.thing.property_action(self))

    async def validate_value(self, value):
        """
        Validate new property value before setting it.
        value -- New value
        """
        if "readOnly" in self.metadata and self.metadata["readOnly"]:
            raise PropertyError("Read-only property")

        try:
            validate(value, self.metadata)
        except ValidationError:
            raise PropertyError("Invalid property value")

    async def as_property_description(self):
        """
        Get the property description.
        Returns a dictionary describing the property.
        """
        description = deepcopy(self.metadata)

        if "links" not in description:
            description["links"] = []

        if description["@type"] in ["VideoProperty", "ImageProperty"]:
            description["links"].append(
                {"rel": "property", "href": self.href, "mediaType": self.media_type}
            )
        else:
            description["links"].append(
                {"rel": "property", "href": self.href_prefix + self.href, "mediaType": self.media_type}
            )
        return description

    async def set_href_prefix(self, prefix):
        """
        Set the prefix of any hrefs associated with this property.
        prefix -- the prefix
        """
        self.href_prefix = prefix

    async def set_href(self, href):
        """
        Set the href associated with this property.
        href -- the new prefix
        """
        self.href = href

    async def set_media_type(self, media_type):
        """
        Set the media type associated with this property.
        media_type -- the new media type
        """
        self.media_type = media_type

    async def get_href(self):
        """
        Get the href of this property.
        Returns the href.
        """
        return self.href_prefix + self.href

    async def get_value(self):
        """
        Get the current property value.
        Returns the value.
        """
        return await self.value.get()

    async def set_value(self, value, with_action=True):
        """
        Set the current value of the property.
        value -- the value to set
        """
        await self.validate_value(value)
        await self.value.set(value, with_action=with_action)

    async def get_name(self):
        """
        Get the name of this property.
        Returns the name.
        """
        return self.name

    async def set_thing(self, thing):
        """Set the thing associated with this property."""
        self.thing = thing

    async def get_thing(self):
        """Get the thing associated with this property."""
        return self.thing

    async def get_metadata(self):
        """Get the metadata associated with this property."""
        return self.metadata



import asyncio

from websockets import ConnectionClosedOK
from starlette import status
from starlette.types import Scope, Receive, Send
from starlette.requests import Request
from starlette.websockets import WebSocket, WebSocketDisconnect
from starlette.responses import UJSONResponse
from starlette.exceptions import HTTPException
from starlette.endpoints import HTTPEndpoint, WebSocketEndpoint

from .auth import requires
from .errors import PropertyError


async def perform_action(action):
    """Perform an Action in a coroutine."""
    await action.start()


class BaseHandler(HTTPEndpoint):
    """Base handler that is initialized with a thing."""

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope, receive, send)
        request = Request(self.scope, receive=self.receive)
        self.things = request.app.state.things

    async def get_thing(self, thing_id):
        """
        Get the thing this request is for.
        things -- list of Things managed by this server
        thing_id -- ID of the thing to get, in string form
        Returns the thing, or None if not found.
        """
        return await self.things.get_thing(thing_id)


class ThingsHandler(BaseHandler):
    """Handle a request to / when the server manages multiple things."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        property_name -- the name of the property from the URL path
        """
        ws_href = f"{'wss' if request.url.scheme == 'https' else 'ws'}://{request.headers.get('Host', '')}"

        descriptions = []
        for idx, thing in await self.things.get_things():
            if not request.app.state.require_auth:
                description = await thing.as_thing_description()
                description["href"] = await thing.get_href()
                description["links"].append({
                    "rel": "alternate",
                    "href": f"{ws_href}{await thing.get_href()}",
                })
                description[
                    "base"
                ] = f"{request.url.scheme}://{request.headers.get('Host', '')}{await thing.get_href()}"

                description["securityDefinitions"] = {
                    "nosec_sc": {"scheme": "nosec", },
                }
                description["security"] = "nosec_sc"
                descriptions.append(description)
            elif request.user.display_name in await thing.get_owners():
                description = await thing.as_thing_description()
                description["href"] = await thing.get_href()
                description["links"].append({
                    "rel": "alternate",
                    "href": f"{ws_href}{await thing.get_href()}",
                })
                description[
                    "base"
                ] = f"{request.url.scheme}://{request.headers.get('Host', '')}{await thing.get_href()}"

                description["securityDefinitions"] = {
                    "bearer_sc": {"scheme": "bearer", },
                }
                description["security"] = "bearer_sc"
                descriptions.append(description)

        return UJSONResponse(descriptions)


class ThingHandler(BaseHandler):
    """Handle a request to /."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request, including websocket requests.
        thing_id -- ID of the thing this request is for
        """
        thing_id = request.path_params.get("thing_id", "0")
        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        ws_href = f"{'wss' if request.url.scheme == 'https' else 'ws'}://{request.headers.get('Host', '')}"

        description = await thing.as_thing_description()
        description["links"].append(
            {"rel": "alternate", "href": f"{ws_href}{await thing.get_href()}", }
        )
        description["base"] = f"{request.url.scheme}://{request.headers.get('Host', '')}{await thing.get_href()}"
        if request.app.state.require_auth:
            description["securityDefinitions"] = {
                "bearer_sc": {"scheme": "bearer", },
            }
            description["security"] = "bearer_sc"
        else:
            description["securityDefinitions"] = {
                "nosec_sc": {"scheme": "nosec", },
            }
            description["security"] = "nosec_sc"

        return UJSONResponse(description)


class WsThingHandler(WebSocketEndpoint):
    """Handle a request to /."""

    encoding = "json"

    async def get_thing(self, thing_id):
        """
        Get the thing this request is for.
        things -- list of Things managed by this server
        thing_id -- ID of the thing to get, in string form
        Returns the thing, or None if not found.
        """
        websocket = WebSocket(self.scope, receive=self.receive, send=self.send)
        things = websocket.app.state.things
        return await things.get_thing(thing_id)

    async def dispatch(self) -> None:
        websocket = WebSocket(self.scope, receive=self.receive, send=self.send)
        try:
            await self.on_connect(websocket)
        except WebSocketDisconnect:
            return

        close_code = status.WS_1000_NORMAL_CLOSURE

        try:
            while True:
                message = await websocket.receive()
                if message["type"] == "websocket.receive":
                    data = await self.decode(websocket, message)
                    await self.on_receive(websocket, data)
                elif message["type"] == "websocket.disconnect":
                    close_code = int(message.get("code", status.WS_1000_NORMAL_CLOSURE))
                    break
        except Exception as exc:
            close_code = status.WS_1011_INTERNAL_ERROR
            raise exc from None
        finally:
            await self.on_disconnect(websocket, close_code)

    @requires('authenticated')
    async def on_connect(self, websocket):
        """
        Handle a GET request, including websocket requests.
        thing_id -- ID of the thing this request is for
        """
        await websocket.accept()

        thing_id = websocket.path_params.get("thing_id", "0")
        self.thing = await self.get_thing(thing_id)

        if self.thing is None:
            try:
                await websocket.send_json(
                    {
                        "messageType": "error",
                        "data": {
                            "status": "404 Not Found",
                            "message": "Invalid thing_id",
                        },
                    },
                    mode="binary",
                )
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

        else:
            await self.thing.add_subscriber(websocket)
            ws_href = f"{websocket.url.scheme}://{websocket.headers.get('Host', '')}"

            description = await self.thing.as_thing_description()
            description["links"].append(
                {"rel": "alternate", "href": f"{ws_href}{await self.thing.get_href()}", }
            )
            description[
                "base"] = f"{websocket.url.scheme}://{websocket.headers.get('Host', '')}{await self.thing.get_href()}"
            if websocket.app.state.require_auth:
                description["securityDefinitions"] = {
                    "bearer_sc": {"scheme": "bearer", },
                }
                description["security"] = "bearer_sc"
            else:
                description["securityDefinitions"] = {
                    "nosec_sc": {"scheme": "nosec", },
                }
                description["security"] = "nosec_sc"

            try:
                await websocket.send_json(description, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def on_receive(self, websocket, message):
        """
        Handle an incoming message.
        message -- message to handle
        """
        if "messageType" not in message or "data" not in message:
            try:
                await websocket.send_json(
                    {
                        "messageType": "error",
                        "data": {
                            "status": "400 Bad Request",
                            "message": "Invalid message",
                        },
                    },
                    mode="binary",
                )
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

        if self.thing is None:
            try:
                await websocket.send_json(
                    {
                        "messageType": "error",
                        "data": {
                            "status": "404 Not Found",
                            "message": "Invalid thing_id",
                        },
                    },
                    mode="binary",
                )
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass
            return

        msg_type = message["messageType"]
        if msg_type == "setProperty":
            for property_name, property_value in message["data"].items():
                try:
                    await self.thing.set_property(property_name, property_value)
                except PropertyError as e:
                    try:
                        await websocket.send_json(
                            {
                                "messageType": "error",
                                "data": {"status": "400 Bad Request", "message": str(e), },
                            },
                            mode="binary",
                        )
                    except (WebSocketDisconnect, ConnectionClosedOK):
                        pass

        elif msg_type == "syncProperty":
            for property_name, property_value in message["data"].items():
                try:
                    await self.thing.sync_property(property_name, property_value)
                except PropertyError as e:
                    try:
                        await websocket.send_json(
                            {
                                "messageType": "error",
                                "data": {"status": "400 Bad Request", "message": str(e), },
                            },
                            mode="binary",
                        )
                    except (WebSocketDisconnect, ConnectionClosedOK):
                        pass

        elif msg_type == "requestAction":
            for action_name, action_params in message["data"].items():
                input_ = None
                if "input" in action_params:
                    input_ = action_params["input"]

                action = await self.thing.perform_action(action_name, input_)
                if action:
                    asyncio.create_task(perform_action(action))
                else:
                    try:
                        await websocket.send_json(
                            {
                                "messageType": "error",
                                "data": {
                                    "status": "400 Bad Request",
                                    "message": "Invalid action request",
                                    "request": message,
                                },
                            },
                            mode="binary",
                        )
                    except (WebSocketDisconnect, ConnectionClosedOK):
                        pass

        elif msg_type == "addEventSubscription":
            for event_name in message["data"].keys():
                await self.thing.add_event_subscriber(event_name, websocket)
        else:
            try:
                await websocket.send_json(
                    {
                        "messageType": "error",
                        "data": {
                            "status": "400 Bad Request",
                            "message": "Unknown messageType: " + msg_type,
                            "request": message,
                        },
                    },
                    mode="binary",
                )
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def on_disconnect(self, websocket, close_code):
        """Handle a close event on the socket."""

        if hasattr(self, "thing") and self.thing:
            await self.thing.remove_subscriber(websocket)


class PropertiesHandler(BaseHandler):
    """Handle a request to /properties."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        """
        thing_id = request.path_params.get("thing_id", "0")
        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(await thing.get_properties())


class PropertyHandler(BaseHandler):
    """Handle a request to /properties/<property>."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        property_name -- the name of the property from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        property_name = request.path_params.get("property_name", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        if await thing.has_property(property_name):
            return UJSONResponse(
                {property_name: await thing.get_property(property_name), }
            )
        else:
            raise HTTPException(status_code=404)

    @requires('authenticated')
    async def put(self, request: Request):
        """
        Handle a PUT request.
        thing_id -- ID of the thing this request is for
        property_name -- the name of the property from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        property_name = request.path_params.get("property_name", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        try:
            args = await request.json()
        except ValueError:
            raise HTTPException(status_code=400)

        if property_name not in args:
            raise HTTPException(status_code=400)

        if await thing.has_property(property_name):
            try:
                await thing.set_property(property_name, args[property_name])
            except PropertyError:
                raise HTTPException(status_code=400)

            return UJSONResponse(
                {property_name: await thing.get_property(property_name), }
            )
        else:
            raise HTTPException(status_code=404)


class ActionsHandler(BaseHandler):
    """Handle a request to /actions."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        """
        thing_id = request.path_params.get("thing_id", "0")

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(await thing.get_action_descriptions())

    @requires('authenticated')
    async def post(self, request: Request):
        """
        Handle a POST request.
        thing_id -- ID of the thing this request is for
        """
        thing_id = request.path_params.get("thing_id", "0")

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        try:
            message = await request.json()
        except ValueError:
            raise HTTPException(status_code=404)

        response = {}
        for action_name, action_params in message.items():
            input_ = None
            if "input" in action_params:
                input_ = action_params["input"]
            print(f"input_ {input_}")
            action = await thing.perform_action(action_name, input_)
            if action:
                response.update(await action.as_action_description())

                # Start the action
                asyncio.create_task(perform_action(action))

        return UJSONResponse(response, status_code=201)


class ActionHandler(BaseHandler):
    """Handle a request to /actions/<action_name>."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        action_name -- name of the action from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        action_name = request.path_params.get("action_name", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(
            await thing.get_action_descriptions(action_name=action_name)
        )

    @requires('authenticated')
    async def post(self, request: Request):
        """
        Handle a POST request.
        thing_id -- ID of the thing this request is for
        action_name -- name of the action from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        action_name = request.path_params.get("action_name", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        try:
            message = await request.json()
        except ValueError:
            raise HTTPException(status_code=404)

        response = {}
        for name, action_params in message.items():
            if name != action_name:
                continue

            input_ = None
            if "input" in action_params:
                input_ = action_params["input"]

            action = await thing.perform_action(name, input_)
            if action:
                response.update(await action.as_action_description())

                # Start the action
                asyncio.create_task(perform_action(action))

        return UJSONResponse(response, status_code=201)


class ActionIDHandler(BaseHandler):
    """Handle a request to /actions/<action_name>/<action_id>."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        action_name -- name of the action from the URL path
        action_id -- the action ID from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        action_name = request.path_params.get("action_name", None)
        action_id = request.path_params.get("action_id", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        action = await thing.get_action(action_name, action_id)
        if action is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(await action.as_action_description())

    @requires('authenticated')
    async def put(self, request: Request):
        """
        Handle a PUT request.
        TODO: this is not yet defined in the spec
        thing_id -- ID of the thing this request is for
        action_name -- name of the action from the URL path
        action_id -- the action ID from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        action_name = request.path_params.get("action_name", None)
        action_id = request.path_params.get("action_id", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse({"msg": "success"}, status_code=200)

    @requires('authenticated')
    async def delete(self, request: Request):
        """
        Handle a DELETE request.
        thing_id -- ID of the thing this request is for
        action_name -- name of the action from the URL path
        action_id -- the action ID from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        action_name = request.path_params.get("action_name", None)
        action_id = request.path_params.get("action_id", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        if await thing.remove_action(action_name, action_id):
            return UJSONResponse({"msg": "success"}, status_code=204)
        else:
            raise HTTPException(status_code=404)


class EventsHandler(BaseHandler):
    """Handle a request to /events."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        """
        thing_id = request.path_params.get("thing_id", "0")

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(await thing.get_event_descriptions())


class EventHandler(BaseHandler):
    """Handle a request to /events/<event_name>."""

    @requires('authenticated')
    async def get(self, request: Request):
        """
        Handle a GET request.
        thing_id -- ID of the thing this request is for
        event_name -- name of the event from the URL path
        """
        thing_id = request.path_params.get("thing_id", "0")
        event_name = request.path_params.get("event_name", None)

        thing = await self.get_thing(thing_id)
        if thing is None:
            raise HTTPException(status_code=404)

        return UJSONResponse(await thing.get_event_descriptions(event_name=event_name))



"""An observable, settable value interface."""

from pyee import AsyncIOEventEmitter as EventEmitter


class Value(EventEmitter):
    """
    A property value.
    This is used for communicating between the Thing representation and the
    actual physical thing implementation.
    Notifies all observers when the underlying value changes through an
    external update (command to turn the light off) or if the underlying sensor
    reports a new value.
    """

    def __init__(self, initial_value, value_forwarder=None):
        """
        Initialize the object.
        initial_value -- the initial value
        value_forwarder -- the method that updates the actual value on the
                           thing
        """
        EventEmitter.__init__(self)
        self.last_value = initial_value
        self.value_forwarder = value_forwarder

    async def set(self, value, with_action=True):
        """
        Set a new value for this thing.
        value -- value to set
        with_action -- do property action
        """
        if self.value_forwarder is not None:
            self.value_forwarder(value)

        await self.notify_of_external_update(value, with_action=with_action)

    async def get(self):
        """Return the last known value from the underlying thing."""
        return self.last_value

    async def notify_of_external_update(self, value, with_action=True):
        """
        Notify observers of a new value.
        value -- new value
        """
        if value is not None:
            self.last_value = value
            if with_action:
                self.emit('update', value)
            else:
                self.emit('sync', value)



from tortoise.models import Model
from tortoise import fields


class Thing(Model):
    # Defining `id` field is optional, it will be defined automatically
    # if you haven't done it yourself
    # id = fields.IntField(pk=True)
    uid = fields.CharField(max_length=255, unique=True)
    title = fields.CharField(max_length=255)

    # Defining ``__str__`` is also optional, but gives you pretty
    # represent of model in debugger and interpreter
    def __str__(self):
        return self.name


# class Rule(Model):
#     name = fields.CharField(max_length=255)
#     description = fields.TextField()



from .event import ThingPairedEvent, ThingRemovedEvent


class SingleThing:
    """A container for a single thing."""

    def __init__(self, thing):
        """
        Initialize the container.
        thing -- the thing to store
        """
        self.thing = thing

    async def get_thing(self, _=None):
        """Get the thing at the given index."""
        return self.thing

    async def get_things(self):
        """Get the list of things."""
        return [self.thing]

    async def get_name(self):
        """Get the mDNS server name."""
        return self.thing.title


class MultipleThings:
    """A container for multiple things."""

    def __init__(self, things, name):
        """
        Initialize the container.
        things -- the things to store
        name -- the mDNS server name
        """
        self.things = things
        self.name = name

    async def get_thing(self, idx):
        """
        Get the thing at the given index.
        idx -- the index
        """
        return self.things.get(idx, None)

    async def get_things(self):
        """Get the list of things."""
        return self.things.items()

    async def get_name(self):
        """Get the mDNS server name."""
        return self.name

    async def add_thing(self, thing):
        self.things.update({thing.id: thing})
        server = self.things.get('urn:webthing:server')

        await server.add_event(ThingPairedEvent({
            '@type': list(thing._type),
            'id': thing.id,
            'title': thing.title
        }))

    async def remove_thing(self, thing_id):
        del self.things[thing_id]
        server = self.things.get('urn:webthing:server')

        await server.add_event(ThingRemovedEvent({
            'id': thing_id,
        }))



"""High-level Event base class implementation."""
from .utils import timestamp


class Event:
    """An Event represents an individual event from a thing."""

    name = None

    def __init__(self, data=None):
        """
        Initialize the object.
        thing -- Thing this event belongs to
        name -- name of the event
        data -- data associated with the event
        """
        self.thing = None
        self.data = data
        self.time = timestamp()

    async def as_event_description(self):
        """
        Get the event description.
        Returns a dictionary describing the event.
        """
        description = {
            self.name: {"timestamp": self.time,},
        }

        if self.data is not None:
            description[self.name]["data"] = self.data

        return description

    async def get_thing(self):
        """Get the thing associated with this event."""
        return self.thing

    async def set_thing(self, thing):
        """Set the thing associated with this event."""
        self.thing = thing

    async def get_name(self):
        """Get the event's name."""
        return self.name

    async def get_data(self):
        """Get the event's data."""
        return self.data

    async def get_time(self):
        """Get the event's timestamp."""
        return self.time


class ThingPairingEvent(Event):
    name = "thing_pairing"


class ThingPairedEvent(Event):
    name = "thing_paired"


class ThingRemovedEvent(Event):
    name = "thing_removed"



from tortoise import Tortoise


async def init():
    # Here we create a SQLite DB using file "db.sqlite3"
    #  also specify the app name of "models"
    #  which contain models from "app.models"
    await Tortoise.init(
        db_url='sqlite://db.sqlite3',
        modules={'models': ['webthing.model']}
    )
    # Generate the schema
    await Tortoise.generate_schemas()



"""High-level Thing base class implementation."""

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from websockets import ConnectionClosedOK
from starlette.websockets import WebSocketDisconnect

from .model import Thing as ThingModel
from .action import Rename
from .event import ThingPairingEvent, ThingPairedEvent, ThingRemovedEvent
from .value import Value
from .property import Property


class Thing:
    """A Web Thing."""

    type = []
    description = ""

    def __init__(self, id_, title, type_=[], description_=""):
        """
        Initialize the object.
        id_ -- the thing's unique ID - must be a URI
        title -- the thing's title
        type_ -- the thing's type(s)
        owners_ -- the thing's owner(s)
        description -- description of the thing
        """
        self._type = set()
        if not isinstance(type_, list):
            self._type.add(type_)
        else:
            self._type = self._type.union(set(type_))

        self._type = self._type.union(set(self.type))

        self.description = description_

        self.id = id_
        self.context = "https://iot.mozilla.org/schemas"
        self.title = title
        self.properties = {}
        self.available_actions = {}
        self.available_events = {}
        self.actions = {}
        self.events = []
        self.subscribers = {}
        self.owners = []
        self.href_prefix = ""
        self.ui_href = None
        self.add_init_action(
            {
                "title": "rename",
                "description": "rename the thing's title",
                "input": {
                    "type": "object",
                    "required": ["title", ],
                    "properties": {
                        "title": {
                            "type": "string",
                        },
                    },
                },
            },
            Rename
        )

    async def as_thing_description(self):
        """
        Return the thing state as a Thing Description.
        Returns the state as a dictionary.
        """
        maybe_thing = await ThingModel.get_or_none(uid=self.id)
        if maybe_thing:
            self.title = maybe_thing.title

        thing = {
            "id": self.id,
            "title": self.title,
            "@context": self.context,
            "properties": await self.get_property_descriptions(),
            "actions": {},
            "events": {},
            "links": [
                {"rel": "properties", "href": f"{self.href_prefix}/properties", },
                {"rel": "actions", "href": f"{self.href_prefix}/actions", },
                {"rel": "events", "href": f"{self.href_prefix}/events", },
            ],
        }

        for name, action in self.available_actions.items():
            thing["actions"][name] = action["metadata"]
            thing["actions"][name]["links"] = [
                {"rel": "action", "href": f"{self.href_prefix}/actions/{name}", },
            ]

        for name, event in self.available_events.items():
            thing["events"][name] = event["metadata"]
            thing["events"][name]["links"] = [
                {"rel": "event", "href": f"{self.href_prefix}/events/{name}", },
            ]

        if self.ui_href is not None:
            thing["links"].append(
                {"rel": "alternate", "mediaType": "text/html", "href": self.ui_href, }
            )

        if self.description:
            thing["description"] = self.description

        if self._type:
            thing["@type"] = list(self._type)

        return thing

    async def get_href(self):
        """Get this thing's href."""
        if self.href_prefix:
            return self.href_prefix

        return "/"

    async def get_ui_href(self):
        """Get the UI href."""
        return self.ui_href

    async def set_href_prefix(self, prefix):
        """
        Set the prefix of any hrefs associated with this thing.
        prefix -- the prefix
        """
        self.href_prefix = prefix

        for property_ in self.properties.values():
            await property_.set_href_prefix(prefix)

        for action_name in self.actions.keys():
            for action in self.actions[action_name]:
                await action.set_href_prefix(prefix)

    async def set_ui_href(self, href):
        """
        Set the href of this thing's custom UI.
        href -- the href
        """
        self.ui_href = href

    async def get_id(self):
        """
        Get the ID of the thing.
        Returns the ID as a string.
        """
        return self.id

    async def get_title(self):
        """
        Get the title of the thing.
        Returns the title as a string.
        """
        return self.title

    async def get_context(self):
        """
        Get the type context of the thing.
        Returns the context as a string.
        """
        return self.context

    async def get_type(self):
        """
        Get the type(s) of the thing.
        Returns the list of types.
        """
        return self.type

    async def get_description(self):
        """
        Get the description of the thing.
        Returns the description as a string.
        """
        return self.description

    async def get_property_descriptions(self):
        """
        Get the thing's properties as a dictionary.
        Returns the properties as a dictionary, i.e. name -> description.
        """
        return {
            k: await v.as_property_description() for k, v in self.properties.items()
        }

    async def get_action_descriptions(self, action_name=None):
        """
        Get the thing's actions as an array.
        action_name -- Optional action name to get descriptions for
        Returns the action descriptions.
        """
        descriptions = []

        if action_name is None:
            for name in self.actions:
                for action in self.actions[name]:
                    descriptions.append(await action.as_action_description())
        elif action_name in self.actions:
            for action in self.actions[action_name]:
                descriptions.append(await action.as_action_description())

        return descriptions

    async def get_event_descriptions(self, event_name=None):
        """
        Get the thing's events as an array.
        event_name -- Optional event name to get descriptions for
        Returns the event descriptions.
        """
        if event_name is None:
            return [await e.as_event_description() for e in self.events]
        else:
            return [
                await e.as_event_description()
                for e in self.events
                if await e.get_name() == event_name
            ]

    async def add_property(self, property_):
        """
        Add a property to this thing.
        property_ -- property to add
        """
        await property_.set_href_prefix(self.href_prefix)
        await property_.set_thing(self)
        self.properties[property_.name] = property_

    async def remove_property(self, property_):
        """
        Remove a property from this thing.
        property_ -- property to remove
        """
        if property_.name in self.properties:
            del self.properties[property_.name]

    async def find_property(self, property_name):
        """
        Find a property by name.
        property_name -- the property to find
        Returns a Property object, if found, else None.
        """
        return self.properties.get(property_name, None)

    async def get_property(self, property_name):
        """
        Get a property's value.
        property_name -- the property to get the value of
        Returns the properties value, if found, else None.
        """
        prop = await self.find_property(property_name)
        if prop:
            return await prop.get_value()

        return None

    async def get_properties(self):
        """
        Get a mapping of all properties and their values.
        Returns a dictionary of property_name -> value.
        """
        return {
            await prop.get_name(): await prop.get_value()
            for prop in self.properties.values()
        }

    async def has_property(self, property_name):
        """
        Determine whether or not this thing has a given property.
        property_name -- the property to look for
        Returns a boolean, indicating whether or not the thing has the
        property.
        """
        return property_name in self.properties

    async def set_property(self, property_name, value):
        """
        Set a property value.
        property_name -- name of the property to set
        value -- value to set
        """
        prop = await self.find_property(property_name)
        if not prop:
            return
        print(f"set {self.id}'s property {property_name} to {value}")
        await prop.set_value(value)

    async def sync_property(self, property_name, value):
        """
        Sync a property value from cloud or mqtt etc.
        property_name -- name of the property to set
        value -- value to set
        """
        prop = await self.find_property(property_name)
        if not prop:
            return
        print(f"sync {self.title}'s property {property_name} to {value}")
        await prop.set_value(value, with_action=False)

    async def get_action(self, action_name, action_id):
        """
        Get an action.
        action_name -- name of the action
        action_id -- ID of the action
        Returns the requested action if found, else None.
        """
        if action_name not in self.actions:
            return None

        for action in self.actions[action_name]:
            if action.id == action_id:
                return action

        return None

    async def add_event(self, event):
        """
        Add a new event and notify subscribers.
        event -- the event that occurred
        """
        self.events.append(event)
        await event.set_thing(self)
        await self.event_notify(event)

    async def add_available_event(self, cls, metadata):
        """
        Add an available event.
        name -- name of the event
        metadata -- event metadata, i.e. type, description, etc., as a dict
        """
        if metadata is None:
            metadata = {}

        self.available_events[cls.name] = {
            "metadata": metadata,
            "subscribers": {},
        }

    async def perform_action(self, action_name, input_=None):
        """
        Perform an action on the thing.
        action_name -- name of the action
        input_ -- any action inputs
        Returns the action that was created.
        """
        if action_name not in self.available_actions:
            return None

        action_type = self.available_actions[action_name]

        if "input" in action_type["metadata"]:
            try:
                validate(input_, action_type["metadata"]["input"])
            except ValidationError:
                return None

        action = action_type["class"](self, input_=input_)
        await action.set_href_prefix(self.href_prefix)
        await self.action_notify(action)
        self.actions[action_name].append(action)
        return action

    async def remove_action(self, action_name, action_id):
        """
        Remove an existing action.
        action_name -- name of the action
        action_id -- ID of the action
        Returns a boolean indicating the presence of the action.
        """
        action = await self.get_action(action_name, action_id)
        if action is None:
            return False

        await action.cancel()
        self.actions[action_name].remove(action)
        return True

    async def add_available_action(self, metadata, cls):
        """
        Add an available action.
        name -- name of the action, default use cls.name
        metadata -- action metadata, i.e. type, description, etc., as a dict
        cls -- class to instantiate for this action
        """
        if metadata is None:
            metadata = {}

        name = cls.name
        self.available_actions[name] = {
            "metadata": metadata,
            "class": cls,
        }
        self.actions[name] = []

    def add_init_action(self, metadata, cls):
        """
        Add an available action.
        name -- name of the action, default use cls.name
        metadata -- action metadata, i.e. type, description, etc., as a dict
        cls -- class to instantiate for this action
        """
        if metadata is None:
            metadata = {}

        name = cls.name
        self.available_actions[name] = {
            "metadata": metadata,
            "class": cls,
        }
        self.actions[name] = []

    async def add_subscriber(self, ws):
        """
        Add a new websocket subscriber.
        ws -- the websocket
        """
        if id(ws) not in self.subscribers:
            self.subscribers[id(ws)] = ws

    async def remove_subscriber(self, ws):
        """
        Remove a websocket subscriber.
        ws -- the websocket
        """
        if id(ws) in self.subscribers:
            self.subscribers.pop(id(ws))

        for name in self.available_events:
            await self.remove_event_subscriber(name, ws)

    async def add_event_subscriber(self, name, ws):
        """
        Add a new websocket subscriber to an event.
        name -- name of the event
        ws -- the websocket
        """
        if name in self.available_events:
            if id(ws) not in self.available_events[name]["subscribers"]:
                self.available_events[name]["subscribers"][id(ws)] = ws

    async def remove_event_subscriber(self, name, ws):
        """
        Remove a websocket subscriber from an event.
        name -- name of the event
        ws -- the websocket
        """
        if (
                name in self.available_events
                and id(ws) in self.available_events[name]["subscribers"]
        ):
            self.available_events[name]["subscribers"].pop(id(ws))

    async def property_notify(self, property_):
        """
        Notify all subscribers of a property change.
        property_ -- the property that changed
        """
        message = {
            "messageType": "propertyStatus",
            "data": {property_.name: await property_.get_value(), },
        }

        for subscriber in list(self.subscribers.values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def property_action(self, property_):
        """
        Addional action when a property change.
        property_ -- the property that changed
        """
        pass

    async def action_notify(self, action):
        """
        Notify all subscribers of an action status change.
        action -- the action whose status changed
        """
        message = {
            "messageType": "actionStatus",
            "data": await action.as_action_description(),
        }

        for subscriber in list(self.subscribers.values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def event_notify(self, event):
        """
        Notify all subscribers of an event.
        event -- the event that occurred
        """

        if event.name not in self.available_events:
            return

        message = {
            "messageType": "event",
            "data": await event.as_event_description(),
        }

        for subscriber in list(self.available_events[event.name]["subscribers"].values()):
            try:
                await subscriber.send_json(message, mode="binary")
            except (WebSocketDisconnect, ConnectionClosedOK):
                pass

    async def add_owner(self, owner: str):
        """
        Add a new owner.
        owner -- the owner
        """
        self.owners.append(owner)
        # await event.set_thing(self)
        # await self.event_notify(event)

    async def get_owners(self):
        """Get this thing's owner(s)."""
        return self.owners


class Server(Thing):
    type = ["Server"]
    description = "Web Thing Environment"

    def __init__(self):
        super().__init__(
            "urn:webthing:server",
            "Web Thing Environment",
        )

    async def build(self):
        await self.add_property(
            Property(
                "state",
                Value("ON"),
                metadata={
                    "@type": "ServerStateProperty",
                    "title": "State",
                    "type": "string",
                    "enum": ["ON", "OFF", "REBOOT"],
                    "description": "state of webthing server",
                },
            )
        )

        await self.add_available_event(
            ThingPairedEvent,
            {
                "description": "new thing paired",
                "type": "object",
                "required": ["@type", "id", "title"],
                "properties": {
                    "@type": {
                        "type": "array",
                    },
                    "id": {
                        "type": "string",
                    },
                    "title": {
                        "type": "string",
                    },
                },
            }
        )

        await self.add_available_event(
            ThingRemovedEvent,
            {
                "description": "device removed event",
                "type": "object",
                "required": ["id", ],
                "properties": {
                    "id": {
                        "type": "string",
                    },
                },
            }
        )

        await self.add_available_event(
            ThingPairingEvent,
            {
                "description": "thing pairing event",
                "type": "object",
                "required": ["id", ],
                "properties": {
                    "id": {
                        "type": "string",
                    },
                },
            }
        )

        return self
